//==========================================================================
//
//      devs_eth_arm_board.inl
//
//      Board ethernet I/O definitions.
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//===========================================================================

/* Changelog
 * Support for ethernet SMSC91Cx modified by Sony
 * Created by original author
 */

#include <cyg/hal/hal_intr.h>           // CYGNUM_HAL_INTERRUPT_ETHR
#include <cyg/hal/hal_if.h>

#ifdef CYGPKG_REDBOOT
#include <pkgconf/redboot.h>
#ifdef CYGSEM_REDBOOT_FLASH_CONFIG
#include <redboot.h>
#include <flash_config.h>
#endif
#endif

extern unsigned int sys_ver;

#ifdef __WANT_DEVS

#ifdef CYGPKG_DEVS_ETH_ARM_MXCBOARD_ETH0
#define LAN91CXX_IS_LAN91C111
#if defined(CYGPKG_REDBOOT) && defined(CYGSEM_REDBOOT_FLASH_CONFIG)
RedBoot_config_option("Set " CYGDAT_DEVS_ETH_ARM_MXCBOARD_ETH0_NAME " network hardware address [MAC]",
                      eth0_esa,
                      ALWAYS_ENABLED, true,
                      CONFIG_BOOL, false
                     );
RedBoot_config_option(CYGDAT_DEVS_ETH_ARM_MXCBOARD_ETH0_NAME " network hardware address [MAC]",
                      eth0_esa_data,
                      "eth0_esa", true,
                      CONFIG_ESA, 0
                     );
#endif // CYGPKG_REDBOOT && CYGSEM_REDBOOT_FLASH_CONFIG

#ifdef CYGSEM_HAL_VIRTUAL_VECTOR_SUPPORT
// Note that this section *is* active in an application, outside RedBoot,
// where the above section is not included.

#include <cyg/hal/hal_if.h>

#ifndef CONFIG_ESA
#define CONFIG_ESA (6)
#endif
#ifndef CONFIG_BOOL
#define CONFIG_BOOL (1)
#endif

cyg_bool _board_provide_eth0_esa(unsigned char * mac)
{
    cyg_bool set_esa;
    int ok;
    ok = CYGACC_CALL_IF_FLASH_CFG_OP( CYGNUM_CALL_IF_FLASH_CFG_GET,
                                      "eth0_esa", &set_esa, CONFIG_BOOL);
    if (ok && set_esa) {
        ok = CYGACC_CALL_IF_FLASH_CFG_OP( CYGNUM_CALL_IF_FLASH_CFG_GET,
                                          "eth0_esa_data", mac, CONFIG_ESA);
    }

    return ok && set_esa;
}
#endif // CYGSEM_HAL_VIRTUAL_VECTOR_SUPPORT

// ------------------------------------------------------------------------
// EEPROM access functions
//
#define PP_ECR                  0x0040
#define PP_EE_READ_CMD          0x0200
#define PP_EE_WRITE_CMD         0x0100
#define PP_EE_EWEN_CMD          0x00F0
#define PP_EE_EWDS_CMD          0x0000
#define PP_EE_ERASE_CMD         0x0300

#define PP_EE_DATA              0x0042
#define PP_EE_ADDR_W0           0x001C
#define PP_EE_ADDR_W1           0x001D
#define PP_EE_ADDR_W2           0x001E

#define EE_TIMEOUT              50000

extern struct eth_drv_sc *__local_enet_sc;

__inline__ void read_mac_addr(unsigned char* buf)
{
       unsigned char   data[6];
       int                             i;
       struct eth_drv_sc *sc = __local_enet_sc;

    (sc->funs->control)(sc, ETH_DRV_GET_MAC_ADDRESS, data, sizeof(data));

       for (i = 0; i < 6; i++) {
               buf[i] = data[i];
       }
}

/*
 * Write a word to an EEPROM location
 * base: package page base (IO base)
 * offset: the EEPROM word offset starting from 0. So for word 1, should pass in 1
 * data: 16 bit data to be written into EEPRM
 */
__inline__ void write_mac_addr(unsigned char* data)
{

    struct eth_drv_sc *sc = __local_enet_sc;
    (sc->funs->control)(sc, ETH_DRV_SET_MAC_ADDRESS, data, sizeof(data));

}

static lan91cxx_priv_data lan91cxx_eth0_priv_data = {
    base : (unsigned short *) 0xB6000300,
#ifdef CYGSEM_DEVS_ETH_ARM_MXCBOARD_ETH0_SET_ESA
    enaddr        : CYGDAT_DEVS_ETH_ARM_MXCBOARD_ETH0_ESA,
    hardwired_esa : true,
#else
    hardwired_esa : false,
#endif
#ifdef CYGSEM_HAL_VIRTUAL_VECTOR_SUPPORT
    provide_esa   : &_board_provide_eth0_esa,
#else
    provide_esa   : NULL,
#endif
};

ETH_DRV_SC(lan91cxx_sc,
           &lan91cxx_eth0_priv_data,          // Driver specific data
           CYGDAT_DEVS_ETH_ARM_MXCBOARD_ETH0_NAME, // Name for device
           lan91cxx_start,
           lan91cxx_stop,
           lan91cxx_control,
           lan91cxx_can_send,
           lan91cxx_send,
           lan91cxx_recv,
           lan91cxx_deliver,
           lan91cxx_poll,
           lan91cxx_int_vector
);

NETDEVTAB_ENTRY(lan91cxx_netdev,
                "lan91cxx_" CYGDAT_DEVS_ETH_ARM_MXCBOARD_ETH0_NAME,
                smsc_lan91cxx_init,
                &lan91cxx_sc);


#endif // CYGPKG_DEVS_ETH_ARM_MXCBOARD_ETH0

#endif // __WANT_DEVS
