
// ========================================================================== 
//                                                                           
//   mxcmci_mmc.c                                                                
//   (c) 2008, Freescale                                                    
//                                                                            
//   MMC card driver for MXC platform                                         
//                                                                           
// ========================================================================== 
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Lewis Liu <weizhi.liu@freescale.com>
// Contributors: Lewis Liu <weizhi.liu@freescale.com>
// Date:         2008-05-13 Initial version
// Purpose:
// Description:
//
//####DESCRIPTIONEND####
//
//==========================================================================


#include <cyg/io/mxcmci_host.h>
#include <cyg/io/mxcmci_core.h>
#include <cyg/io/mxcmci_mmc.h>
#include <cyg/io/mxc_mmc.h>

static cyg_uint32 mmc_set_rca (void);

cyg_uint32 address_mode; /* Global variable for addressing mode */


cyg_uint32 mmc_init(cyg_uint32 bus_width)
{
    cyg_uint32 status = (cyg_uint32)(-1);
    cyg_uint32 spec_version;
    /* Get CID number of MMC Card */
    if(!mxcmci_get_cid())
    {
        /* Set RCA of the MMC Card */
        if(!mmc_set_rca())
        {
            flash_dprintf(FLASH_DEBUG_MAX,"%s:  mmc_set_rca OK!", __FUNCTION__);
            /* Get Spec version supported by the card */
            spec_version = mmc_get_spec_ver();

            /*Enable operating frequency */
            host_cfg_clock(OPERATING_FREQ);

            /*Put MMC in Transfer State */
            if(!mxcmci_trans_prepare ())
            {
            	flash_dprintf(FLASH_DEBUG_MAX,"%s:  mxcmci_trans_prepare OK!", __FUNCTION__);
                /* Set status variable as SUCESS */
                status = 0; 
             }

         }
    }
 
    return status;
	
}

cyg_uint32 mmc_data_read (cyg_uint32 *dest_ptr,cyg_uint32 length,cyg_uint32 offset)
{
    command_t cmd;
    int len;
    cyg_uint32 read_block_status=0;
    cyg_uint32 blk_len = BLK_LEN;
	
    /* Assing length of data to be read */
    len = length;
   
    if(address_mode == SECT_MODE)  
    {
        blk_len = (BLK_LEN/512);
        offset =(offset/BLK_LEN);
    }
    
    /* Configure interface block and number of blocks */
    host_cfg_block(BLK_LEN,ONE);
 
    /* Configure CMD16 to set block length as 512 bytes.*/
    mxcmci_cmd_config(&cmd,CMD16,BLK_LEN,READ,RESPONSE_48,
                    DATA_PRESENT_NONE,ENABLE,ENABLE);
	
    /* Issue command CMD16 to set block length as 512 bytes */
    if(host_send_cmd(&cmd) == FAIL)
    {
    	flash_dprintf(FLASH_DEBUG_MAX,"%s: Can't set block length!\n", __FUNCTION__);
    	esdhc_softreset(ESDHC_RESET_CMD_MSK);
        read_block_status = FAIL;
    }
    else
    {
        while(len>0 && !read_block_status) 
        {
            
            /* Comfigure command CMD17 for single block read */	
            mxcmci_cmd_config(&cmd,CMD17,offset,READ,RESPONSE_48,
	                        DATA_PRESENT,ENABLE,ENABLE);	

			//available_mask = ESDHCI_DATA_AVAILABLE;
            if(host_send_cmd(&cmd) == FAIL)
            {
            flash_dprintf(FLASH_DEBUG_MAX,"%s: Can't send CMD17!\n", __FUNCTION__);
				esdhc_softreset(ESDHC_RESET_CMD_MSK | ESDHC_RESET_DAT_MSK);
                read_block_status = FAIL;
            }
            else
            {
                /* Call interface Data read function */
                read_block_status = host_data_read(dest_ptr,BLK_LEN);
				
                /*decrement length by block size */
                len = len - BLK_LEN;
				
                /* Increment offset by Block Size */
                offset = offset + blk_len;

               /* Increment Destination pointer */
               dest_ptr = dest_ptr + (BLK_LEN/FOUR);
            }
	
        }
				
    }
    return read_block_status;
}


cyg_uint32 mmc_data_write (cyg_uint32 *dest_ptr,cyg_uint32 length,cyg_uint32 offset)
{

	command_t cmd;
    int len;
    cyg_uint32 write_block_status=SUCCESS;
	
    /* Assing length of data to be read */
    len = length;
    
    if(address_mode == SECT_MODE)  
    {
        offset /= BLK_LEN;
    }
     /* Configure interface block and number of blocks */
     host_cfg_block(BLK_LEN,ONE);
 
    /* Configure CMD16 to set block length as 512 bytes.*/
    mxcmci_cmd_config(&cmd,CMD16,BLK_LEN,READ,RESPONSE_48,
                    DATA_PRESENT_NONE,ENABLE,ENABLE);
	
    /* Issue command CMD16 to set block length as 512 bytes */
    if(host_send_cmd(&cmd) == FAIL)
    {
        flash_dprintf(FLASH_DEBUG_MAX,"%s: can't set block length\n", __FUNCTION__);
		esdhc_softreset(ESDHC_RESET_CMD_MSK);
        write_block_status = FAIL;
    }
    else
    {
        while(len>0 && !write_block_status) 
        {
			
            /* Comfigure command CMD24 for single block write */	
            mxcmci_cmd_config(&cmd,CMD24,offset,WRITE,RESPONSE_48,
	                        DATA_PRESENT,ENABLE,ENABLE);	

			   //available_mask = ESDHCI_SPACE_AVAILABLE;
            if(host_send_cmd(&cmd) == FAIL)
            {
				
                flash_dprintf(FLASH_DEBUG_MAX,"%s: Failed in configuring CMD24\n", __FUNCTION__);
				    esdhc_softreset(ESDHC_RESET_CMD_MSK | ESDHC_RESET_DAT_MSK);
                write_block_status = FAIL;
            }
            else
            {
               /* Call interface write read function */
               write_block_status = host_data_write(dest_ptr,BLK_LEN);
               //flash_dprintf(FLASH_DEBUG_MAX,"%s: write_block_status=%d\n", __FUNCTION__, write_block_status);
				
               /*decrement length by block size */
               len = len - BLK_LEN;
				
               /* Increment offset by Block Size */
		         address_mode == SECT_MODE ? (offset++) : (offset +=  BLK_LEN);

               /* Increment Destination pointer */
               dest_ptr = dest_ptr + (BLK_LEN/FOUR);
            }
	
        }
				
    }
    return write_block_status;

}


cyg_uint32 mmc_data_erase (cyg_uint32 offset, cyg_uint32 size)
{
	command_t cmd;
	cyg_uint8 startEraseBlockCmd = CMD35; 
	cyg_uint8 endEraseBlockCmd   = CMD36;
	cyg_uint32 startBlock = offset/BLK_LEN; 
	cyg_uint32 endBlock = (offset + size - 1)/BLK_LEN;
	cyg_uint32 ret;

   //flash_dprintf(FLASH_DEBUG_MAX,"address_mode=%d\n", address_mode);
	if(address_mode != SECT_MODE) {
        	startBlock *= BLK_LEN;
		endBlock *= BLK_LEN;
		startEraseBlockCmd = CMD32; 
		endEraseBlockCmd   = CMD33;
    	}
	
	/* Configure start erase command to set first block*/
	mxcmci_cmd_config(&cmd,startEraseBlockCmd,startBlock,READ,RESPONSE_48,
                    DATA_PRESENT_NONE,ENABLE,ENABLE);
    	/* wait response */
	if((ret = host_send_cmd(&cmd)) == SUCCESS){
      flash_dprintf(FLASH_DEBUG_MAX,"%s: successful for host_send_cmd\n", __FUNCTION__);
    	/* Configure end erase command to set end block*/
		mxcmci_cmd_config(&cmd,endEraseBlockCmd,endBlock,READ,RESPONSE_48,
	        		              DATA_PRESENT_NONE,ENABLE,ENABLE);
	   if((ret = host_send_cmd(&cmd)) == SUCCESS){
         flash_dprintf(FLASH_DEBUG_MAX,"%s: successful for host_send_cmd:2\n", __FUNCTION__);
		   /* Comfigure command to start erase*/	
         mxcmci_cmd_config(&cmd,CMD38,0,READ,RESPONSE_48,
	                              DATA_PRESENT_NONE,ENABLE,ENABLE);
	      if((ret = host_send_cmd(&cmd)) == SUCCESS){
            flash_dprintf(FLASH_DEBUG_MAX,"%s: successful for host_send_cmd:3\n", __FUNCTION__);
	         //wait for completion
		      return ret;             	
         }
	   }
	}
	
   flash_dprintf(FLASH_DEBUG_MAX,"%s: Error return (%d)\n", __FUNCTION__, ret);
	return ret;
}


cyg_uint32 mmc_voltage_validation (void)
{
    command_t cmd;
    command_response_t response;
    cyg_uint32 voltage_validation_command=0;
    cyg_uint32 ocr_val=0;
    cyg_uint32 voltage_validation = FAIL;
	
    ocr_val = (cyg_uint32) ((MMC_OCR_VALUE) & 0xFFFFFFFF);
	
    while((voltage_validation_command < MMCSD_READY_TIMEOUT ) && (voltage_validation != SUCCESS)) 
    {  	
        /* Configure CMD1 for MMC card */
        mxcmci_cmd_config(&cmd, CMD1, ocr_val, READ, RESPONSE_48,
                                  DATA_PRESENT_NONE, DISABLE, DISABLE);
                 
        /* Issue CMD1 to MMC card to determine OCR value */
        if(host_send_cmd(&cmd) == FAIL)
        {
             voltage_validation = FAIL;
             break;
        }
        else
        {
            /* Read Response from CMDRSP0 Register */
            response.format = RESPONSE_48;
            host_read_response(&response);
		
            /* Check if card busy bit is cleared or not */
            if(!(response.cmd_rsp0 & CARD_BUSY_BIT))
            {
                /* Iterate One more time */
                voltage_validation_command++;
            }
            else
            {
                if((response.cmd_rsp0 & MMC_OCR_HC_RES) == MMC_OCR_HC_RES)
                {  
                    address_mode = SECT_MODE; 
                    voltage_validation = SUCCESS;
                }
                else if((response.cmd_rsp0 & MMC_OCR_LC_RES) == MMC_OCR_LC_RES)
                {
                    address_mode = BYTE_MODE; 
                    voltage_validation = SUCCESS;
                }
            }
                     
        }
    }
	
    return voltage_validation;
}

static cyg_uint32 mmc_set_rca (void)
{
    command_t cmd;
    cyg_uint32 card_state=0;
    cyg_uint32 rca_request=0;
    command_response_t response;
    cyg_uint32 card_address = (Card_rca << RCA_SHIFT);
	
    /* Configure CMD3 for MMC card */
    /* 32bit card address is expected as Argument */
    mxcmci_cmd_config(&cmd,CMD3,card_address,READ,RESPONSE_48,
                    DATA_PRESENT_NONE,ENABLE,ENABLE);
	                    
	                    
    /* Assigns relative address to the card
    */
     
    if(host_send_cmd(&cmd) == FAIL)
    {
        rca_request = FAIL;
    } 
	
    else
    {
        /* Read Command response */
        response.format = RESPONSE_48;
        host_read_response (&response);
        card_state = CURR_CARD_STATE(response.cmd_rsp0);
	     if(card_state == IDENT)
        {
            rca_request = SUCCESS;
			
	     }
        else
        {
            rca_request = FAIL;
        }
    }
		
    return rca_request;
}

cyg_uint32 mmc_get_spec_ver (void)
{
		
    cyg_uint32 mmc_spec_version;

	if(card_get_csd() == FAIL)
		{
		mmc_spec_version = 0;
    }
	else
		{
		mmc_spec_version = ((csd.csd3 && MMC_SPEC_VER) >> MMC_SPEC_VER_SHIFT); 
	}
	
    return mmc_spec_version;
		
}

cyg_uint32 card_flash_query(void* data)
{
    command_t cmd;
    cyg_uint32 cid_request = FAIL;
    command_response_t response;

    /* Configure CMD2 for card */
    mxcmci_cmd_config(&cmd, CMD2, NO_ARG, READ, RESPONSE_136, DATA_PRESENT_NONE, ENABLE, DISABLE);
    /* Issue CMD2 to card to determine CID contents */
    if(host_send_cmd(&cmd) == FAIL)
    {
        cid_request = FAIL;
        flash_dprintf(FLASH_DEBUG_MAX,"%s: can't send query command\n", __FUNCTION__);
    }
    else
    {
        cyg_uint32* d = (cyg_uint32*)data;
        /* Read Command response  */
        response.format= RESPONSE_136;
        host_read_response (&response);
        /* Assign CID values to mmc_cid structures */
        *d++ = response.cmd_rsp0;
        *d++ = response.cmd_rsp1;
        *d++= response.cmd_rsp2;
        *d= response.cmd_rsp3;

        /* Assign cid_request as SUCCESS */
        cid_request = SUCCESS;
    }
    flash_dprintf(FLASH_DEBUG_MAX, "%s(Success?=%d):(ID=0x%x: 0x%x, 0x%x, 0x%x)\n", 
                  __FUNCTION__, cid_request,*(cyg_uint32*)(data), *(cyg_uint32*)((cyg_uint32)data+4),
                  *(cyg_uint8*)((cyg_uint32)data+8), *(cyg_uint8*)((cyg_uint32)data+12));
    return;
}
