// ========================================================================== 
//                                                                           
//   mxcmci_sd.c                                                                
//   (c) 2008, Freescale                                                    
//                                                                            
//   MMC card driver for MXC platform                                         
//                                                                           
// ========================================================================== 
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Lewis Liu <weizhi.liu@freescale.com>
// Contributors: Lewis Liu <weizhi.liu@freescale.com>
// Date:         2008-05-13 Initial version
// Purpose:
// Description:
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <cyg/io/mxcmci_host.h>
#include <cyg/io/mxcmci_core.h>
#include <cyg/io/mxcmci_mmc.h>

static cyg_uint32 sd_get_rca (void);
static cyg_uint32 sd_get_bit_mode_support(void);
static cyg_uint32 sd_set_bus_width (cyg_uint32);
static cyg_uint32 sd_set_boot_partition (void);                                

#define SD_OCR_VALUE_HV_LC 0x00ff8000 /* nirp_oct07: <- 3.3v, LC */
#define SD_OCR_VALUE_HV_HC 0x40ff8000 /* nirp_oct07: <- 3.3v, HC */
/* nirp_oct07: LV_LC not needed - 1.8v is only supported under eSD which supports HC by default (SD>2.00) */
#define SD_OCR_VALUE_LV_HC 0x40000080 /* nirp_oct07: <- 1.8v, HC */

#define SD_OCR_HC_RES 0x40000000
#define SD_OCR_LC_RES 0x00000000


#define SD_IF_HV_COND_ARG 0x000001AA
#define SD_IF_LV_COND_ARG 0x000002AA

#define RCA_SHIFT 16
#define SD_R1_STATUS_APP_CMD_MSK 0x20
#define BIT_MODE_4_SUPPORT 5
#define SD_BUS_WIDTH_OFFSET 6
#define BIT_4_MODE 4
#define SD_STATUS_LEN 64

#define SD_BOOT_SWITCH_ARG 0x80FFFF2F
#define SD_PARTITION1 0x01000000

cyg_uint32 sd_init(cyg_uint32 bus_width)
{
    cyg_uint32 status = FAIL;
    cyg_uint32 bus_size = bus_width; 
	
    /* Get CID number of SD Memory Card */
    if(!mxcmci_get_cid ())
    {
    	//diag_printf("%s:mxcmci_get_cid OK!\n", __FUNCTION__);
        /* Set RCA of the SD Card */
        if(!sd_get_rca())
        {
        	//diag_printf("%s:sd_get_rca OK!\n", __FUNCTION__);
            /*Enable operating frequency */
            host_cfg_clock(OPERATING_FREQ);

            /*Get CSD from Card*/
            card_get_csd();
            /*Put SD Card in Transfer State */
            if(!mxcmci_trans_prepare ())
            {
            	//diag_printf("%s:mxcmci_trans_prepare OK!\n", __FUNCTION__);
                /* Set Boot Partition to partition #1 */
                sd_set_boot_partition();

                status = SUCCESS; 
             }
        }
    }
	
	//diag_printf("%s:failed to Init SD card!\n", __FUNCTION__);
    return status;
	
}

cyg_uint32 sd_voltage_validation (void)
{
	//wait max timeout (unit: ms)
	cyg_uint32 timeout = 15000;	

    command_t cmd;
    command_response_t response;
    cyg_uint32 voltage_validation_command =0;
    cyg_uint32 default_rca = 0;
    
	
    cyg_uint32 ocr_value = SD_OCR_VALUE_HV_LC; /* nirp_oct07: <- split OCR to 3.3v and 1.8v cases */
    cyg_uint32 voltage_validation = FAIL;
    cyg_uint32 interface_value = 0;
    cyg_uint32 card_usable = SUCCESS;

    /* Configure Command CMD8 to check for High capacity support */
    mxcmci_cmd_config(&cmd,CMD8,SD_IF_HV_COND_ARG,READ,RESPONSE_48,DATA_PRESENT_NONE,ENABLE,ENABLE);

    /* Issue Command CMD8  to SD Memory card */
    if(host_send_cmd(&cmd) == SUCCESS) /* nirp_oct07: <- changed order of detection */
    {
            //diag_printf("%s:CMD8 OK!\n", __FUNCTION__);
        response.format =  RESPONSE_48;
        host_read_response(&response);

        /* Obtain Interface value from the response buffer */
        interface_value = response.cmd_rsp0;
        
        /* Check if volatge lies in range or not*/
        if((interface_value & SD_IF_HV_COND_ARG) == SD_IF_HV_COND_ARG)
        {
            ocr_value = ((cyg_uint32)(SD_OCR_VALUE_HV_HC) & 0xFFFFFFFF); /* nirp_oct07: <- split OCR to 3.3v and 1.8v cases */
        }
        
        /* start timer for a  delay of 1.5sec, for ACMD41 */
    	//GPT_delay(ESDHC_ACMD41_TIMEOUT);
    	timeout = 15000;
   
    	while((voltage_validation_command < MMCSD_READY_TIMEOUT) && (voltage_validation != SUCCESS) && 
         (card_usable == SUCCESS)  )
        {
	        /* Configure CMD55 for SD card */
	        /* This command expects defualt RCA 0x0000 as argument.*/
	        mxcmci_cmd_config(&cmd,CMD55,default_rca,READ,RESPONSE_48,
	                                  DATA_PRESENT_NONE,ENABLE,ENABLE);
		                    
		           
	        /* Issue CMD55 to SD Memory card*/
	        if(host_send_cmd(&cmd) == FAIL)
	        {
	            voltage_validation = FAIL;
	            break;
	        }
	        else
	        {
				/* Configure ACMD41 for SD card */
				/* This command expects operating voltage range as argument.*/
				/* CODE REVIEW START: Need to check why BUSY was expected */	
				/* INTERNAL CODE REVIEW: Accepted - to fix original code if needed */
				/* nirp: changed RESPONSE_48_CHECK_BUSY to RESPONSE_48 */
	           /* nirp_oct03: why with busy again? ACMD41 doesn't hold busy line */
	            mxcmci_cmd_config(&cmd,ACMD41,ocr_value,READ,RESPONSE_48,DATA_PRESENT_NONE,DISABLE,DISABLE);
			
				/* CODE REVIEW END:  */
			
	            /* Issue ACMD41 to SD Memory card to determine OCR value */
	            if(host_send_cmd(&cmd) == FAIL)
	            {
	                voltage_validation = FAIL;
	                break;
	            }
	            else
	            {
	                /* Read Response from CMDRSP0 Register */
	                response.format = RESPONSE_48;
	                host_read_response(&response);
				
	                /* Obtain OCR Values from the response */
	                /* Obtain OCR value from the response buffer
	                */
	                ocr_value = response.cmd_rsp0;

	                /* Check if card busy bit is cleared or not */
	                if(!(response.cmd_rsp0 & CARD_BUSY_BIT))
	                {
	                    /* Iterate One more time */
	                    voltage_validation_command++;
	                }
	                else
	                {

						/*CODE REVIEW START: Update code and check only bit 30, HC or LC card type. All voltage bits needs to be masked. */
						/* INTERNAL CODE REVIEW: Accepted - need fix the code accordingly */
						/* nirp: It may be better to check the actual power supply voltage - requiring the entire range (0xff8000) may fail the sequence even if the device can be supported */
						/*CODE REVIEW END: */

	                                      
	                    if((response.cmd_rsp0 & SD_OCR_HC_RES) == SD_OCR_HC_RES)
	                    {
	                        address_mode = SECT_MODE;
	                        voltage_validation = SUCCESS;
	                    }
	                    /* CODE REVIEW 3: (same as above) Check is logically correct, but seems redundent. 
	                    Anything that fails the HC check, is assumed Low Capacity */
	                    /* nirp_oct03: this can be just an "else". the LC macro is 0 anyway, 
	                    and anything not HC is LC by default */
	                    /* removed else if*/ 
	                    else 
	                    {
	                        address_mode = BYTE_MODE;
	                        voltage_validation = SUCCESS;
	                     }
	                }
	            }

	        }
			if( --timeout == 0){
				break;
				}
			hal_delay_us(1000);
		  
    	}
         
        if (voltage_validation == FAIL)
        {
	         card_usable = FAIL; 
		}
           
    }
    else
    {
         //diag_printf("%s:3.3v test failed, try to test 1.8v mode!\n", __FUNCTION__);
         mxcmci_cmd_config(&cmd,CMD8,SD_IF_LV_COND_ARG,READ,RESPONSE_48,DATA_PRESENT_NONE,ENABLE,ENABLE);
        
              
	     /* Issue Command CMD8  to SD Memory card */
	    if(host_send_cmd(&cmd) == FAIL)
		{
		//diag_printf("%s:CMD8 for 1.8v failed!\n", __FUNCTION__);
			/* nirp_oct07: CMD8 failed both in 3.3 and in 1.8v, try SD 1.x case - no CMD8, LC, 3.3v only */
            ocr_value = ((cyg_uint32)(SD_OCR_VALUE_HV_LC) & 0xFFFFFFFF);  /* nirp_oct07: <- changed order of detection */
		}
		else
	    {
	         		//diag_printf("%s:CMD8 for 1.8v OK!\n", __FUNCTION__);

	        response.format =  RESPONSE_48;
	        host_read_response(&response);

	        /* Obtain Interface value from the response buffer */
	        interface_value = response.cmd_rsp0;
	        
	        /* Check if volatge lies in range or not*/
	        if((interface_value & SD_IF_LV_COND_ARG) == SD_IF_LV_COND_ARG)
	        {
	            ocr_value = ((cyg_uint32)(SD_OCR_VALUE_LV_HC) & 0xFFFFFFFF); /* nirp_oct07: <- split OCR to 3.3v and 1.8v cases */
	        }
			/* nirp_oct07: otherwise, try with HV_LC settings (set at function start) */
	                  
	    }
    
    }

  /* start timer for a  delay of 1.5sec, for ACMD41 */
    //GPT_delay(ESDHC_ACMD41_TIMEOUT);
    timeout = 15000;
   
   
   /* nirp_oct03: MMCSD_READY_TIMEOUT too long. 
   ACMD41 also takes longer than CMD1 (twice - ~200 clocks for CMD55+resp+CMD41+resp */
   /* In any case ,ACMD 41 will loop not more than 1.5 sec*/ 
    while((voltage_validation_command < MMCSD_READY_TIMEOUT) && (voltage_validation != SUCCESS) && 
         (card_usable == SUCCESS)) 
    {
        /* Configure CMD55 for SD card */
        /* This command expects defualt RCA 0x0000 as argument.*/
        mxcmci_cmd_config(&cmd,CMD55,default_rca,READ,RESPONSE_48,
                                  DATA_PRESENT_NONE,ENABLE,ENABLE);
	                    
	           
        /* Issue CMD55 to SD Memory card*/
        if(host_send_cmd(&cmd) == FAIL)
        {
            voltage_validation = FAIL;
					//diag_printf("%s:CMD55 failed!\n", __FUNCTION__);

            break;
        }
        else
        {
			/* Configure ACMD41 for SD card */
			/* This command expects operating voltage range as argument.*/
			/* CODE REVIEW START: Need to check why BUSY was expected */	
			/* INTERNAL CODE REVIEW: Accepted - to fix original code if needed */
			/* nirp: changed RESPONSE_48_CHECK_BUSY to RESPONSE_48 */
           /* nirp_oct03: why with busy again? ACMD41 doesn't hold busy line */
            mxcmci_cmd_config(&cmd,ACMD41,ocr_value,READ,RESPONSE_48,DATA_PRESENT_NONE,DISABLE,DISABLE);
		
			/* CODE REVIEW END:  */
		
            /* Issue ACMD41 to SD Memory card to determine OCR value */
            if(host_send_cmd(&cmd) == FAIL)
            {
                voltage_validation = FAIL;
				//diag_printf("%s:ACMD41 failed!\n", __FUNCTION__);
                break;
            }
            else
            {
                /* Read Response from CMDRSP0 Register */
                response.format = RESPONSE_48;
                host_read_response(&response);
			
                /* Obtain OCR Values from the response */
                /* Obtain OCR value from the response buffer
                */
                ocr_value = response.cmd_rsp0;

                /* Check if card busy bit is cleared or not */
                if(!(response.cmd_rsp0 & CARD_BUSY_BIT))
                {
                    /* Iterate One more time */
                    voltage_validation_command++;
                }
                else
                {

					/*CODE REVIEW START: Update code and check only bit 30, HC or LC card type. All voltage bits needs to be masked. */
					/* INTERNAL CODE REVIEW: Accepted - need fix the code accordingly */
					/* nirp: It may be better to check the actual power supply voltage - requiring the entire range (0xff8000) may fail the sequence even if the device can be supported */
					/*CODE REVIEW END: */

                                      
                    if((response.cmd_rsp0 & SD_OCR_HC_RES) == SD_OCR_HC_RES)
                    {
                        address_mode = SECT_MODE;
                        voltage_validation = SUCCESS;
                    }
                    /* CODE REVIEW 3: (same as above) Check is logically correct, but seems redundent. 
	                Anything that fails the HC check, is assumed Low Capacity */
	                /* nirp_oct03: this can be just an "else". the LC macro is 0 anyway, 
	                and anything not HC is LC by default */
                    else 
                    {
                        address_mode = BYTE_MODE;
                        voltage_validation = SUCCESS;
                     }
                }
            }

        }
    if( --timeout == 0){
				break;
				}
			hal_delay_us(1000);
    
    }

	
    return voltage_validation;
}

static cyg_uint32 sd_get_rca (void)
{
    command_t cmd;
    cyg_uint32 card_state=0;
    cyg_uint32 rca_request=0;
    command_response_t response;
	
    /* Configure CMD3 for MMC card */
    /* 32bit card address is expected as Argument */
    mxcmci_cmd_config(&cmd,CMD3,NO_ARG,READ,RESPONSE_48,
                    DATA_PRESENT_NONE,ENABLE,ENABLE);
	                             
    /* Get relative address of the card
     */
     
    if(host_send_cmd(&cmd) == FAIL)
    {
        rca_request = FAIL;
    }
    else
    {
        /* Read Command response */
        response.format = RESPONSE_48; 
        host_read_response (&response);
		
        Card_rca = ((cyg_uint32) (response.cmd_rsp0 >> RCA_SHIFT));
		
        card_state = CURR_CARD_STATE(response.cmd_rsp0);
	
        if(card_state == IDENT)
        {
            rca_request = SUCCESS;
			
        }
        else		
        {
            rca_request = FAIL;
        }
    }
		
	return rca_request;
}

static cyg_uint32 sd_get_bit_mode_support(void)
{
    command_t cmd;
    cyg_uint32 rd_data_buff[128];
    cyg_uint32 bit4_mode_support;
    command_response_t response;
    cyg_uint32 card_address = (Card_rca << RCA_SHIFT);
	
    /* Configure CMD55 for SD card */
    /* This command expects RCA as argument.*/
    mxcmci_cmd_config(&cmd,CMD55,card_address,READ,RESPONSE_48,
                                  DATA_PRESENT_NONE,ENABLE,ENABLE);
	                                  	           
    /* Issue CMD55 to SD Memory card*/
    if(host_send_cmd(&cmd) == FAIL)
    {
        bit4_mode_support = 0;
    }
    else
    {
        /* Read Command response */
        response.format = RESPONSE_48;
        host_read_response (&response);
		
        /* Afetr giving ACMD Command, the R1 response should have
         * STATUS_APP_CMD set
	 */
        if(response.cmd_rsp0 & SD_R1_STATUS_APP_CMD_MSK)
	{
			
	    /* Configure ACMD51 for SD card */
	    /* This command expects No argument.*/
			
            mxcmci_cmd_config(&cmd,ACMD51,NO_ARG,READ,RESPONSE_48,DATA_PRESENT,ENABLE,ENABLE);
		
            /* Issue ACMD51 to SD Memory card */
            if(host_send_cmd(&cmd) == FAIL)
            {
                bit4_mode_support = 0;
            }
            else
            {
                /* Read Response from e-SDHC buffer */
                host_data_read(rd_data_buff,512);
				
                /* Check for bus width supported */
                bit4_mode_support = (rd_data_buff[SD_BUS_WIDTH_OFFSET] & BIT_MODE_4_SUPPORT);
				
                if(bit4_mode_support)
                {
                    bit4_mode_support = BIT_4_MODE;
                }
				
            }
        }
    }
	
    return bit4_mode_support;
}


static cyg_uint32 sd_set_bus_width (cyg_uint32 bus_width)
{
    command_t cmd;
    cyg_uint32 set_bus_width_status=0;
    command_response_t response;
    cyg_uint32 card_address = (Card_rca << RCA_SHIFT);
	
    if ((bus_width == FOUR ) || (bus_width == ONE))
    {
        /* Configure CMD55 for SD card */
        /* This command expects RCA as argument.*/
		
        mxcmci_cmd_config(&cmd,CMD55,card_address,READ,RESPONSE_48,
	                                  DATA_PRESENT_NONE,ENABLE,ENABLE);
		                    
		           
        /* Issue CMD55 to SD Memory card*/
        if(host_send_cmd(&cmd) == FAIL)
        {
            set_bus_width_status = FAIL;
        }
        else
        {
            /* Read Command response */
            response.format = RESPONSE_48;
            host_read_response (&response);
			
            /* Afetr giving ACMD Command, the R1 response should have
             * STATUS_APP_CMD set
             */
            if(response.cmd_rsp0 & SD_R1_STATUS_APP_CMD_MSK)
            {
			                          
                bus_width = (bus_width>>ONE);
				
                /* Configure ACMD6 for SD card */
                mxcmci_cmd_config(&cmd,ACMD6,bus_width,READ,RESPONSE_48,DATA_PRESENT_NONE,ENABLE,ENABLE);
		        
		       
                /* Issue ACMD6 to SD Memory card*/
                if(host_send_cmd(&cmd) == FAIL)
                {
                    set_bus_width_status = FAIL;
                }
				
                else
                {
                    set_bus_width_status = SUCCESS;
                }
            }
        }
    }
	
    return set_bus_width_status;	
}


/*==========================================================================
FUNCTION: cyg_uint32 sd_set_boot_partition (void)
DESCRIPTION:
  sd_set_boot_partition() will set set boot partition for Partition1 

ARGUMENTS PASSED:
   None

RETURN VALUE:
   SUCCESS / FAILURE

PRE-CONDITIONS:
   None

POST-CONDITIONS:
   None

Detailed Description:

==============================================================================*/

static cyg_uint32 sd_set_boot_partition (void)
{
    command_t cmd;
    cyg_uint32 set_partition_status=FAIL;
    command_response_t response;
    cyg_uint8 response_data[512]; 
    cyg_uint32 *response_pointer = (cyg_uint32 *)response_data;
    cyg_uint32 card_address = (Card_rca << RCA_SHIFT); 
    cyg_uint32 card_state;
    
	/* CODE REVIEW START: Surender to reivew Nir commnents. We will followup to update the eSD function */
    /* INTERNAL CODE REVIEW: Host controller code is written for 512Byte block, 
       Ask Sandisk whether can keep 512 block size */
    /* nirp2: Configure interface block and number of blocks */
     
     /* Jameer :updated according to second review comments ,temp doing block length as 64 byts*/
     host_cfg_block(ESDHC_CONFIG_BLOCK_64,ONE);  /* nirp2: added block configuration */
     
     
     /* Configure CMD6 to switch the command */
     mxcmci_cmd_config(&cmd,CMD6,SD_BOOT_SWITCH_ARG,READ,RESPONSE_48,DATA_PRESENT,ENABLE,ENABLE);

    if(host_send_cmd(&cmd) == FAIL)
    {
            set_partition_status = FAIL;
    }
    else
    {
        /* Read 512 bit data strcuture returned from command CMD6 */
        /* jameer:updated according to review comments */ 
        host_data_read(response_pointer, SWITCH_BLK_LEN); 
            
        /* nirp: changed BLK_LEN to SWITCH_BLK_LEN (64B) */
	 	/* INTERNAL CODE REVIEW: Host controller code is written for 512Byte block, */  
        /* 	Ask Sandisk whether can keep 512 block size */
        /* check for bits [383:380], it must contain the value 0x2 */
		/* nirp: why response_data[95] ??? */
		/* nirp: todo - calculate actual address */
    	/* INTERNAL CODE REVIEW: Surrender says t should be bit [47], 
    	need to check with Sandisk on both 17 Vs 47 and 0x20 vs. 0x02. Need change '&&' to '&' */

		 /* nirp: changed to read byte 17, high nibble */
		/* Byte 17, bits 4-7 */
		/* nirp 9 oct :use [16] */
        if((response_data[16] & 0xF0) == 0x20) 
            {
			/* Send CMD37 to select partition PARTITION1 */
			/* nirp: changed RESPONSE_48 to RESPONSE_48_CHECK_BUSY to wait for busy */
			/* INTERNAL CODE REVIEW: Change code to use _BUSY */
            mxcmci_cmd_config(&cmd,CMD37,SD_PARTITION1,READ,RESPONSE_48_CHECK_BUSY,DATA_PRESENT_NONE,ENABLE,ENABLE);
            if(host_send_cmd(&cmd) == FAIL)
            {
				/* nirp: if CMD6 succeeded switching to function 2, CMD37 MUST be supported - so this can be considered a critical error */
                 set_partition_status = FAIL;
            }
            else
            {
      
				/* Configure CMD13 to read status of the card becuase CMD7 has R1b response */
				/* nirp: CMD13 can report OUT_OF_RANGE after CMD37 in case partition #1 does not exist - this can be used to detect a non bootable eSD device */
				/* INTERNAL CODE REVIEW: Keep code as is, don't modify */
               mxcmci_cmd_config(&cmd,CMD13,card_address,READ,RESPONSE_48,DATA_PRESENT_NONE,ENABLE,ENABLE);

               if(host_send_cmd(&cmd) == FAIL)
               {
                   set_partition_status = FAIL; 
               }
               
               else
               {
                    /* Read Command response */
                    response.format = RESPONSE_48;
                    host_read_response (&response);

                    card_state = CURR_CARD_STATE(response.cmd_rsp0);

                    if(card_state == TRAN)
                    {
                         set_partition_status = SUCCESS;

                    }
                }
            }
         }

         else
         {
             set_partition_status = FAIL;

         }
    }
    
    
     /* Jameer :updated according to second review comments */
     host_cfg_block(ESDHC_CONFIG_BLOCK_512,ONE);  /* nirp2: added block configuration */
     
    
 return set_partition_status;
} 
               

