//==========================================================================
//
//      flash_config.h
//
//      Flash configuration data tables for RedBoot
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    gthomas
// Contributors: gthomas
// Date:         2000-08-21
// Purpose:      
// Description:  
//              
// This code is part of RedBoot (tm).
//
//####DESCRIPTIONEND####
//
//==========================================================================

/* Changelog
 * 2009/11/13 Changes regarding firmware update added by Sony
 * Created by original author
 */

#ifndef _FLASH_CONFIG_H_
#define _FLASH_CONFIG_H_

#define MAX_SCRIPT_LENGTH CYGNUM_REDBOOT_FLASH_SCRIPT_SIZE
#define MAX_STRING_LENGTH CYGNUM_REDBOOT_FLASH_STRING_SIZE
#define MAX_CONFIG_DATA   CYGNUM_REDBOOT_FLASH_CONFIG_SIZE

#define CONFIG_EMPTY   0
#define CONFIG_BOOL    1
#define CONFIG_INT     2
#define CONFIG_STRING  3
#define CONFIG_SCRIPT  4
#ifdef CYGPKG_REDBOOT_NETWORKING
#define CONFIG_IP      5
#define CONFIG_ESA     6
#define CONFIG_NETPORT 7
#endif

struct config_option {
    char *key;
    char *title;
    char *enable;
    bool  enable_sense;
    int   type;
    unsigned long dflt;
} CYG_HAL_TABLE_TYPE;

#define ALWAYS_ENABLED (char *)0

#define RedBoot_config_option(_t_,_n_,_e_,_ie_,_type_,_dflt_)        \
struct config_option _config_option_##_n_                               \
CYG_HAL_TABLE_QUALIFIED_ENTRY(RedBoot_config_options,_n_) =             \
   {#_n_,_t_,_e_,_ie_,_type_,(unsigned long)_dflt_};

// Cause the in-memory configuration data to be written to flash
void flash_write_config(bool prompt);
// Fetch a data item from flash storage, returns 'false' if not found
bool flash_get_config(char *key, void *val, int type);
// Update a data item from flash storage, returns 'false' if not found
bool flash_set_config(char *key, void *val, int type);
// Enumerate keys from configuration
bool flash_next_key(char *key, int keylen, int *type, int *offset);
// Add a new data item to configuration data base.  Returns 'false'
// if no space is available.
bool flash_add_config(struct config_option *opt, bool update);

// Internal structure used to hold config data
struct _config {
    unsigned long len;
    unsigned long key1;
    unsigned char config_data[MAX_CONFIG_DATA-(4*4)];
    unsigned long key2;
    unsigned long cksum;
};

/* define this, if you want the firmware update to be performed */
#define FIRMWARE_UPDATE

#ifdef FIRMWARE_UPDATE

/* if the success and failure is not previously declared, do it now */
#ifndef SUCCESS
#define SUCCESS						(0)
#endif

#ifndef FAILURE
#define FAILURE						(-1)
#endif

/* Flash area related declarations */
#define FLASH_FLAGS_AREA_ADDRESS	(0x100000)
#define FLASH_FLAGS_AREA_READ_SIZE	(0x1000)
#define FLASH_REDBOOT_CONFIG_ADDR	(0x9F000)

#define CONFIG_UPDATE_TO_BUNCH_1	(1)
#define CONFIG_UPDATE_TO_BUNCH_2	(2)
#define CONFIG_CORRUPT_CHK_UPDATE_1 (3)
#define CONFIG_CORRUPT_CHK_UPDATE_2 (4)
#define CONFIG_UPDATE_TO_KERNEL_1	'1'
#define CONFIG_UPDATE_TO_KERNEL_2	'2'

#define CONFIG_UPDATE_TO_FS_1		'6'
#define CONFIG_UPDATE_TO_FS_2		'8'

#define FLASH_POS_STATUS_BYTE       (0)
#define FLASH_POS_VERSION_MAJOR     (1)
#define FLASH_POS_VERSION_MINOR     (2)
#define FLASH_POS_VERSION_MAJOR_OLD (3)
#define FLASH_POS_VERSION_MINOR_OLD (4)

#define STATUS_BYTE_POS_STF			(7)
#define STATUS_BYTE_POS_RESERVED	(6)
#define STATUS_BYTE_POS_NVF			(1)
#define STATUS_BYTE_POS_BVF			(0)

#define SEARCH_FOR_KERNEL_1			"kernel1"
#define SEARCH_FOR_KERNEL_2			"kernel2"
/* Status byte structure */
typedef struct _tag_flash_status_byte
{
    unsigned char stf               : 1;
    unsigned char reserved          : 5;
    unsigned char nvf               : 1;
    unsigned char bvf               : 1;
}flag_status_byte;

/* The status byte is followed by major version and minor version
 * The old_major_version & old_minor_version is required in case of
 * fail-safe mechanism
 */
typedef struct _tag_flash_flag_area
{
    flag_status_byte    status_byte;
    unsigned char       major_version;
    unsigned char       minor_version;
    unsigned char       old_major_version;
    unsigned char       old_minor_version;
}flash_flag_area;

#endif //FIRMWARE_UPDATE

#endif // _FLASH_CONFIG_H_
